package me.lysne.thesis.levert;

import java.awt.*;
import java.awt.image.BufferedImage;

public class ValueNoiseAlgorithm implements DNAToImageTranslator {

    private static final int IMG_SIZE = 1 << 10;
    private static final int CELL_SIZE = 12;

    private double grid2ImageUnits;

    public BufferedImage generateImage(String sequence) {

        double[][][] grid = makeGrid(sequence);
        BufferedImage img = new BufferedImage(IMG_SIZE, IMG_SIZE, BufferedImage.TYPE_INT_RGB);

        grid2ImageUnits = ((IMG_SIZE - 1) / ((grid.length - 1) * 1.));

        for (int y = 0; y < grid.length - 1; y++) {
            for (int x = 0; x < grid.length - 1; x++) {
                expand(x, y, grid, img);
            }
        }

        return img;
    }

    private void expand(int x0, int y0, double[][][] grid, BufferedImage img) {

        double f = 0.0, g = 0.0;
        double gradientFactor = 1. / grid2ImageUnits;
        double[] rgb = new double[3];

        for (int y = 0; y < grid2ImageUnits; y++) {

            int y1 = y0 + 1;

            for (int x = 0; x < grid2ImageUnits; x++) {

                int x1 = x0 + 1;

                double r1 = lerp(grid[y0][x0][0], grid[y0][x1][0], f);
                double g1 = lerp(grid[y0][x0][1], grid[y0][x1][1], f);
                double b1 = lerp(grid[y0][x0][2], grid[y0][x1][2], f);

                double r2 = lerp(grid[y1][x0][0], grid[y1][x1][0], f);
                double g2 = lerp(grid[y1][x0][1], grid[y1][x1][1], f);
                double b2 = lerp(grid[y1][x0][2], grid[y1][x1][2], f);

                rgb[0] = lerp(r1, r2, g);
                rgb[1] = lerp(g1, g2, g);
                rgb[2] = lerp(b1, b2, g);

                img.setRGB(
                        (int) Math.ceil(x0 * grid2ImageUnits + x),
                        (int) Math.ceil(y0 * grid2ImageUnits + y),
                        invertBrightness(Util.doubleToRGB(rgb)));

                f += gradientFactor;
                if (f >= 1.0) f = 0;
            }

            g += gradientFactor;
            if (g >= 1.0) g = 0;
            f = 0;
        }
    }


    private static double[][][] makeGrid(String sequence) {

        int gridSize = (int) Math.ceil(Math.sqrt((double) sequence.length() / CELL_SIZE));

        double[][][] grid = new double[gridSize][gridSize][3];
        double d = 1.0 / CELL_SIZE;
        int i = 0;
        int j = 0;

        for (int k = 0; k < sequence.length(); k++) {

            switch (sequence.charAt(k)) {
                case 'A':
                    grid[i][j][0] += d;
                    break;
                case 'C':
                    grid[i][j][1] += d;
                    break;
                case 'T':
                    grid[i][j][2] += d;
                    break;
                case 'G':
                    grid[i][j][0] += d / 4.0;
                    grid[i][j][1] += d / 4.0;
                    break;
                default: continue;
            }

            if ((k + 1) % CELL_SIZE == 0 && k < (sequence.length() - 1)) {
                j++;
                if (j >= gridSize) {
                    j = 0;
                    i++;
                }
            }
        }
        return grid;
    }

    private static int invertBrightness(int rgb) {
        if (rgb < 0x010101) return rgb;
        int r = (rgb >> 16) & 0xFF;
        int g = (rgb >>  8) & 0xFF;
        int b = (rgb)       & 0xFF;
        float[] hsb = Color.RGBtoHSB(r, g, b, null);
        return Color.HSBtoRGB(hsb[0], hsb[1], 1f - hsb[2]);
    }

    private static double lerp(double a, double b, double f) {
        return (1 - f) * a + f * b;
    }
}
